:- module(analyse,[ analyse_idiom/5]).
:- use_module(normalform).
:- use_module(logging).
:-  use_module(listUtil).


%  A list of basic techniques
% see (Bowles & Brna: Introductory prolog: a suitable
% selection of Programming Techniques ) 


/** 
The same technique. 

it is used to pass the same value between the head of
a clause and a recursive subgoal in the clause.
p(... X ...):-
	...,
	p(...X...)
	...
The same technique appears in the thrid argument of rev/3
rev([],R,R).
rev([H|T],R0,R):-
	rev(T,[H|R0],R).
Our approach to technques is that they capture some declarative
relationship between an argument position in the head of a recursive
clause and the same argument position in the recursive subgoal. Same is
the most simple instance of this where the values in these two argument
positions are the same; that is, the relationship captured is equality.
*/
technique(Arg,same(Value),Clause):-
	recursive_case(Clause,_,RecSubgoal,_),
	Clause=Head-Subgoals,
	member(RecSubgoal,Subgoals),
	arg(Arg,Head,Value),
	arg(Arg,RecSubgoal,Value).
	

/**
The list head technqiue
list head technique is used hwen the head value is the list and the subgoal value is its tail.
p([H|T]):-p(T).
*/
technique(Arg,list_head([A|X]),Clause):-
	recursive_case(Clause,_,RecSubgoal,_),
	Clause=Head-Subgoals,
	member(RecSubgoal,Subgoals),
	arg(Arg,Head,[A|X]),
	arg(Arg,RecSubgoal,X).
/**
The technique list subgoal 
is where the subgoal value is the list and the head value is its tail.
p(T):-p([H|T]).

in "Programming technique for Prolog" Brna gives a name for H
as constructor parameter, T as recursion parameter.
*/
technique(Arg,list_subgoal([X]),Clause):-
	logging(analyse,'<technique_list_subgoal begins>Clause=~w',[Clause]),
	recursive_case(Clause,_,RecSubgoal,_),
	Clause=Head-_Subgoals,
	arg(Arg,Head,X),
	arg(Arg,RecSubgoal,[_|X]),
	logging(analyse,'<technique_list_subgoal ends>Arg=~w, X=~w',[Arg,X]).

/**
The after technique.
nrev([],[]).
nrev([H|T],R):-
	nrev(T,TR),
	append(TR,[H],R).
In this case, there are non recursive subgoals after the recrusive one.
We call this the after technique.
p(X):-
	p(Y),
	g(Y,X)
The value of X is constructed in the subgoal g from the value Y produced
by the recursive call. Note that the relationship between the arguments
in the head of the clause, the recursive subgoal and the subgoal after
the recursive subgoal is an important part of the definition.

Corresponding to this is the before technique.
p(X):-
	g(X,Y),
	p(Y).
This occurs in the follwoing program to compute the length of a list:
length([],L,L).
length([H|T],L0,L):-
	L1 is L0+1,
	length(T,L1,L).
*/



% technique(compare_after(NonRecursive),Clause)
% if this subcall is nonrecursive, occurs after the recursive call and 
% has a functor that is a comparison operator.
technique(after(RecSubgoal,NonRecursive),Clause):-
	logging(analyse,'<after begins>Clause=~w',[Clause]),
	recursive_case(Clause,_,RecSubgoal,NonRecursiveList),
	member(NonRecursive,NonRecursiveList),
	check_position_technique(Clause,RecSubgoal,NonRecursive),!,
	logging(analyse,'<after ends>NonRecursive=~w',[NonRecursive]).
technique(before(NonRecursive,RecSubgoal),Clause):-
	logging(analyse,'<before begins>Clause=~w',[Clause]),
	recursive_case(Clause,NonRecursiveList,RecSubgoal,_),
	member(NonRecursive,NonRecursiveList),
	check_position_technique(Clause,RecSubgoal,NonRecursive),!,
	logging(analyse,'<before ends>NonRecursive=~w',[NonRecursive]).


% technique(compare_after(NonRecursive),Clause)
% if this subcall is nonrecursive, occurs after the recursive call and 
% has a functor that is a builtin predicate (not comparison operator).
/**
technique(builtin_after(RecSubgoal,NonRecursive),Clause):-
	logging(analyse,'<builtin_after begins>Clause=~w',[Clause]),
	recursive_case(Clause,_,RecSubgoal,NonRecursiveList),
	member(NonRecursive,NonRecursiveList),
	check_position_technique(Clause,RecSubgoal,NonRecursive),
	!,
	once(builtin(NonRecursive)),
	logging(analyse,'<builtin_after ends>NonRecursive=~w',[NonRecursive]).

% N is the argument that exists in NonRecursive subgoal.
technique(testcase_before(NonRecursive,RecSubgoal),Clause):-
	logging(analyse,'<compare_before begins>Clause=~w',[Clause]),
	recursive_case(Clause,NonRecursiveList,RecSubgoal,_),
	member(NonRecursive,NonRecursiveList),
	check_position_technique(Clause,RecSubgoal,NonRecursive),
	!,
	functor(NonRecursive,Functor,Arity),
	%logging(analyse,'<compare_before>Functor F1=~w',[F1]),
	once(compare_op(Functor); Arity=1),
	logging(analyse,'<compare_before ends>NonRecursive=~w',[NonRecursive]).
*/
technique(builtin_before(NonRecursive,RecSubgoal),Clause):-
	logging(analyse,'<builtin_before begins>Clause=~w',[Clause]),
	recursive_case(Clause,NonRecursiveList,RecSubgoal,_),
	member(NonRecursive,NonRecursiveList),
	check_position_technique(Clause,RecSubgoal,NonRecursive),
	!,
	once(builtin(NonRecursive)),
	logging(analyse,'<builtin ends>NonRecursive=~w',[NonRecursive]).

%=compare_op(Op)
% Op ist ein Vergleichsoperator fr Zahlen
compare_op(<).
compare_op(>).
compare_op(=<).
compare_op(>=).
compare_op(\=).
compare_op(==).
compare_op(\==).

builtin(member(_,_)).
builtin(append(_,_,_)).


% check_nonrec_position(+Clause,+RecSubgoal,+NonRecursive):-
check_position_technique(Clause,RecSubgoal,NonRecursive):-
	logging(analyse,'<check_pos_technique begins>Clause=~w, RecSubgoal=~w, NonRecursive=~w',[Clause,RecSubgoal,NonRecursive]),
	Clause=Head-_Subgoals,
	Head=..[_F|ArgsHead],
	flatten(ArgsHead,ArgsHeadFlatten),
	member(X,ArgsHeadFlatten),
	%logging(analyse,'<check_pos_technique>one argument in the head X=~w',[X]),
	% the value of X is constructed in the nonrecursive subgoal g from the value Y produced by the recursive call.
	NonRecursive=..[_F1|ArgsNonRec],
	flatten(ArgsNonRec,ArgsNonRecFlatten),
	member(X,ArgsNonRecFlatten),
	%logging(analyse,'<check_pos_technique>found the same argument in the nonrecursive call',[]),
	delete(ArgsNonRecFlatten,X,ArgsNonRecFlatten2),
	RecSubgoal=..[_F2|ArgsRec],
	flatten(ArgsRec,ArgsRecFlatten),
	member(Y,ArgsNonRecFlatten2),
	%logging(analyse,'<check_pos_technique>one argument in the nonrecursive call Y=~w',[Y]),
	
	member(Y,ArgsRecFlatten),
	%logging(analyse,'<check_pos_technique>found the same argument in the recursive call',[]),
	logging(analyse,'<check_pos_technique ends> relationship between argument succesful',[])
	.
	
% --------------------------------------------------------------- %
% a list of schemata 

/**
a clause is a base case if it does not have a body and have a base
constructor.
*/	
% basecase(+Clause,-base_constructor(+Index,-NonvarList))
% Clause is a clause to be tested if it is a base case
% NonvarList is a list of arguments that make the base_constructor.
% IndexList contains the position information of the arguments in the NonvarList.
base_case(Head-[], base_constructor(IndexList,NonvarList)):-
	logging(analyse,'<base_case begins>Input BaseHead=~w',[Head]),
%	BaseCase=Head-[],
	Head=..[_F|Args],
	find_ground(Args,1,IndexList,NonvarList),
	logging(analyse,'<base_case ends>Output IndexList=~w, NonvarList=~w',[IndexList,NonvarList]).

find_ground([Nonvar|Rest],Index,[Index|IndexList],[Nonvar|NonvarList]):-
%	logging(analyse,'<find_ground1 begins>Nonvar=~w, Rest=~w, Index=~w',[Nonvar,Rest,Index]),
	(   Nonvar\='<var>'(_), Nonvar\='<a-var>'(_)),
	!,
	Index1 is Index +1,
	find_ground(Rest,Index1,IndexList,NonvarList),
	logging(analyse,'<find_ground1 ends>',[]).
find_ground([Var|Rest],Index,IndexList,NonvarList):-
%	logging(analyse,'<find_ground2 begins>var=~w, Rest=~w, Index=~w',[Var,Rest,Index]),
	(   Var='<var>'(_); Var='<a-var>'(_)),
	Index1 is Index +1,
	find_ground(Rest,Index1, IndexList,NonvarList),
	logging(analyse,'<find_ground2 ends>',[]).
find_ground([],_,[],[]).


% recursive_case(+Head-NBody,-Before,-Call,-After) 
% Head-NBody is a clause
% Before is a subgoal before a recursive call 
% Call is a recursive subgoal
% After is a subgoal after a recursive call 

recursive_case(Head-NBody,Before,Call,After):-
	logging(analyse,'<recursive_case>Head-NBody=~w-~w',[Head,NBody]),
	functor(Head,Functor,N),
	functor(Call,Functor,N),
	append(Before,[Call|After],NBody),
	logging(analyse,'<recursive_case>This is a recursive case: Before=~w,Call=~w,After=~w',[Before,Call,After]).

%nonrecursive_case(Clause):-
%	not(recursive_case(Clause,_,_,_)). 

% ------------------------------------------------------------------- %
% begin analysing the predicates.

	
% analyse_idiom(+Schema,+StudentPred,+SamplePred,+Index,-ErrorList)
% Schema is a pattern that the matched sample and student solutions applied
% Index is the position of the clause that we are working on
% StudentPred and SamplePred are sorted list of possible matched clauses

analyse_idiom(Schema,[StudentClause|StudentClauseRest],[SampleClause|SampleClauseRest],Index,ErrorList):-
	logging(analyse,'<analyse_idiom1 begins>Technique=~w, StudentClause=~w, StudentClause=~w, Index=~w',[Schema,StudentClause,SampleClause,Index]),
	schema(Schema,StudentClause, SampleClause,Index,SchemaErrors),
	(   SchemaErrors\=[]->
	    analyse_clause(StudentClause,SampleClause,Index,ClauseError);
	    ClauseError=[]
	),
	Index1 is Index+1,
	logging(analyse,'<analyse_pred1>Index1=~w',[Index1]),
	    
	append(SchemaErrors,ClauseError,IdiomErrors),
	append(IdiomErrors,ErrorList1,ErrorList),
	analyse_idiom(Schema,StudentClauseRest,SampleClauseRest,Index1,ErrorList1),
	logging(analyse,'<analyse_idiom1 ends>ErrorList=~w',[ErrorList])
	.
analyse_idiom(Schema,[],[SampleClause|SampleClauseRest],Index,ErrorList):-
	logging(analyse,'<analyse_idiom2 begins>Schema=~w,StudentPred=[], SamplePred=~w|~w, Index=~w',[Schema,SampleClause,SampleClauseRest,Index]),
	%schema(Schema,[], SampleClause,Index,SchemaErrors),
	(   base_case(SampleClause,_)->ClauseErrors=[idiom(error([Index],missing_basecase(SampleClause)))];
	recursive_case(SampleClause,_,_,_)->ClauseErrors=[idiom(error([Index],missing_recursivecase(SampleClause)))]),
	
	Index1 is Index +1,
	logging(analyse,'<analyse_idiom2>Index1=~w',[Index1]),
	%append(SchemaErrors,ClauseErrors,IdiomErrors),
	append(ClauseErrors,ErrorList1,ErrorList),
	analyse_idiom(Schema,[],SampleClauseRest,Index1,ErrorList1),
	logging(analyse,'<analyse_idiom2 ends>ErrorList=~w',[ErrorList])
	.
analyse_idiom(Schema, [StudentClause|StudentClauseRest],[],Index,ErrorList):-
	logging(analyse,'<analyse_pred3 begins>StudentPred=~w, SamplePred=[], Index=~w',[StudentClause,StudentClauseRest,Index]),
	%schema(Schema,StudentClause,[],Index,SchemaErrors),
	base_case(StudentClause,_)->ClauseErrors=[idiom(error([Index],superfluous_basecase(StudentClause)))];
	recursive_case(StudentClause,_,_,_)->ClauseErrors=[idiom(error([Index],superfluous_recursivecase(StudentClause)))],
        Index1 is Index +1,
	%append(SchemaErrors,ClauseErrors,IdiomErrors),
	append(ClauseErrors,ErrorList1,ErrorList),
	analyse_idiom(Schema,[],StudentClauseRest,Index1,ErrorList1),
	logging(analyse,'<analyse_pred3 ends>ErrorList=~w',[ErrorList])
	.
analyse_idiom(_Schema,[],[],_,[]).

/** 
see Prolog Programming, Paul Brna, Jan 24, 2001, Page 72. 
The test_process is a pattern which guarantees that some inputs will only be processed if the input passes a test.

 test_process(Info,[],[]).
 
 test_process(Info,[Head|Tail],[Head|Result]):-
      test(Info,Head),
      test_process(Info,Tail,Result).
 
 test_process(Info,[Head|Tail],Result):-
      not(test(Info,Head)),
      test_process(Info,Tail,Result).
*/
/**
 a test_process schema has normally 2 subgoals like the pattern above. We can determine the output argument and the Info list of arguments by the following inference. 
The output variable exists in the 2 nd subgoal and in the head, not in the 1.st subgoal. 
The Info list is the rest of argument list of the 2 nd subgoal.
After we've found the Info list, with the 1. st subgoal we can find the the input variable if the 1 st subgoal has more than 2 arguments, otherwise the 1 subgoal has one argument which is the input variable and no Info list.
*/

% schema(test_process,+StudentClause,+SampleClause,+Index,-ErrorList)
% Index ist the index of the clause within the predicate definition.

schema(test_process,StudentClause,SampleClause,Index,ErrorList):-
	logging(analyse,'<schema_test_process begins>StudentClause=~w, SampleClause=~w',[StudentClause,SampleClause]),
	(   
	% we assume that the input clauses have the same types: base case or recursive case.
	% Different types of input clauses will be handled by the analyse_clause
        % if the input clauses are base cases
	% then determine the Info list and compare between the student and sample solution.
	(base_case(StudentClause,base_constructor(BaseIndex1,_)), 
	  base_case(SampleClause,base_constructor(BaseIndex2,_))
	)
	 ->
	(   logging(analyse,'<schema_test_process>both clauses are base case',[]),
	    % determin the Info list of the student solution
	    StudentClause=StudentClauseHead-[],
	    StudentClauseHead=..[_F1|StudentArgs],
	    length(StudentArgs,StudentArgsLength),
	    logging(analyse,'<schema_test_process>StudentArgsLength=~w',[StudentArgsLength]),
	    numlist(1,StudentArgsLength,StudentNumlist),
	    find_rest_list(StudentNumlist,BaseIndex1,StudentInfoList),

	    % determin the Info list of the sample solution
	    SampleClause=SampleClauseHead-[],
	    SampleClauseHead=..[_F2|SampleArgs],
	    length(SampleArgs,SampleArgsLength),
	    numlist(1,SampleArgsLength,SampleNumlist),
	    find_rest_list(SampleNumlist,BaseIndex2,SampleInfoList),
	    (	StudentInfoList\=SampleInfoList -> 
		                 ErrorsBase=[idiom(error([Index], unmatched_info(StudentInfoList,SampleInfoList)))];
				 ErrorsBase=[]
	    ),
	    logging(analyse,'<schema_test_process>ErrorBase=~w',[ErrorsBase]),
	    ErrorList=ErrorsBase						    

	);

	% if the input clauses are recursive cases
	% then determine test subgoal, process subgoal, Info list, input argument and output argument.
	% finally compare these schema elements between the student and sample solution.

	(recursive_case(StudentClause,TestSubgoals1,RecCall1,_),
	 recursive_case(SampleClause,TestSubgoals2,RecCall2,_)
	)
	 ->
	(
	 logging(analyse,'<schema_test_process>both clauses are recursive case',[]), 
	 % an info argument exists in the head, in a test subgoal, in the recursive call
	 StudentClause=StudentHead-_,
	 SampleClause= SampleHead-_,
	 (   TestSubgoals1=[]->true;
	         logging(analyse,'<schema_test_process>TestSubgoals1=~w',[TestSubgoals1]), 
	         member(TestSubgoal1,TestSubgoals1)
	 ),
	         find_info_list_for_test_process(StudentHead,TestSubgoal1,RecCall1,StudentInfoList),

	 (   TestSubgoals2=[]->true;
	        logging(analyse,'<schema_test_process>TestSubgoals2=~w',[TestSubgoals2]), 
	        member(TestSubgoal2,TestSubgoals2)
	 ),   
	        find_info_list_for_test_process(SampleHead,TestSubgoal2,RecCall2,SampleInfoList),

	 % the input argument can be extracted from the argument list of the test subgoal except the info list.
	 % the output argument can be extracted from the argument list of the head except the info list, the input argument.
	 
	 logging(analyse,'<schema_test_process>TestSubgoals1=~w',[TestSubgoals1]),
	 logging(analyse,'<schema_test_process>StudentInfoList=~w',[StudentInfoList]),
	 StudentHead=..[_FH1|StudentHeadArgs],
	 (   TestSubgoals1\=[]->TestSubgoal1=..[_FT1|StudentTestArgs];
	                    StudentTestArgs=[]),
         logging(analyse,'<schema_test_process>StudentHeadArgs=~w',[StudentHeadArgs]),
	 find_input_output_for_test_process(StudentClause,StudentHeadArgs,StudentTestArgs,StudentInfoList,StudentInputList,StudentOutputList),
	 logging(analyse,'<schema_test_process>StudentInputList=~w',[StudentInputList]),
	 logging(analyse,'<schema_test_process>StudentOutputList=~w',[StudentOutputList]),
	 SampleHead=..[_FH2|SampleHeadArgs],
	 (   TestSubgoals2\=[]-> TestSubgoal2=..[_FT2|SampleTestArgs];
	                         SampleTestArgs=[]),       
	 find_input_output_for_test_process(SampleClause,SampleHeadArgs,SampleTestArgs,SampleInfoList,SampleInputList,SampleOutputList),
	 logging(analyse,'<schema_test_process>SampleInputList=~w',[SampleInputList]),
	 logging(analyse,'<schema_test_process>SampleOutputList=~w',[SampleOutputList]),

	 % now compare the test subgoal
	 
	 compareSubgoalList(Index,test_subgoal,TestSubgoals1,TestSubgoals2,ErrorsTest),
	 logging(analyse,'<schema_test_process>ErrorsTest=~w',[ErrorsTest]),
	 
	 % now compare the process subgoal
	 logging(analyse,'<schema_test_process>RecCall1=~w',[RecCall1]), 
	 logging(analyse,'<schema_test_process>RecCall2=~w',[RecCall2]), 
	 compareSubgoalList(Index,process_subgoal,RecCall1,RecCall2,ErrorsProcess),
	 logging(analyse,'<schema_test_process>ErrorsProcess=~w',[ErrorsProcess]),
	 
	 % compare the info list
	 logging(analyse,'<schema_test_process>StudentInfoList=~w',[StudentInfoList]),
	 logging(analyse,'<schema_test_process>SampleInfoList=~w',[SampleInfoList]),
	 compareArgList(Index,info,StudentInfoList,SampleInfoList,ErrorsInfo),
	 logging(analyse,'<schema_test_process>ErrorsInfo=~w',[ErrorsInfo]),
	 
	 % compare the input list
	 logging(analyse,'<schema_test_process>StudentInputList=~w',[StudentInputList]),
	 logging(analyse,'<schema_test_process>SampleInputList=~w',[SampleInputList]),
	 compareArgList(Index,input,StudentInputList,SampleInputList,ErrorsInput),
	 logging(analyse,'<schema_test_process>ErrorsInput=~w',[ErrorsInput]),
	 
	 % compare the output list
	 logging(analyse,'<schema_test_process>StudentOutputList=~w',[StudentOutputList]),
	 logging(analyse,'<schema_test_process>SampleOutputList=~w',[SampleOutputList]),
	 compareArgList(Index,output,StudentOutputList,SampleOutputList,ErrorsOutput),
	 logging(analyse,'<schema_test_process>ErrorsOutput=~w',[ErrorsOutput]),

	 append(ErrorsTest,ErrorsProcess,Errors1),
	 append(Errors1,ErrorsInfo,Errors2),
	 append(Errors2,ErrorsInput,Errors3),
	 append(Errors3,ErrorsOutput,ErrorList)
	 
	)
	),
	logging(analyse,'<schema_test_process ends> ErrorList=~w',[ErrorList])
.

% compareSubgoal(+Index,+SubgoalType,+StudentSubgoal,+SampleSubgoal,-Errors)
% SubgoalType has types: test_subgoal, process_subgoal, recursive_subgoal
compareSubgoalList(Index,SubgoalType,StudentSubgoalList,SampleSubgoalList,Errors):-
	logging(analyse,'<compareSubgoalList1 begins > StudentSubgoal=~p, SampleSubgoal=~p, SubgoalType=~p',[StudentSubgoalList,SampleSubgoalList,SubgoalType]),
	zuordnen(StudentSubgoalList, SampleSubgoalList,ids_state(10,_Pos, [], _State0),_State1)-> Errors=[];
	Errors=[idiom(error([Index], unmatched(SubgoalType,StudentSubgoalList,SampleSubgoalList)))],
	logging(analyse,'<compareSubgoalList1 ends > SubgoalType=~w, Errors=~p',[SubgoalType,Errors])
	.

compareSubgoalList(Index,SubgoalType,StudentSubgoalList,SampleSubgoalList,Errors):-
	logging(analyse,'<compareSubgoalList2 begins> StudentSubgoalList=~p, SampleSubgoalList=~p, SubgoalType=~w',[StudentSubgoalList,SampleSubgoalList,SubgoalType]),
	length(StudentSubgoalList,StudentSubgoalLength),
	length(SampleSubgoalList,SampleSubgoalLength),
	logging(analyse,'<compareSubgoalList2> StudentSubgoalLength=~p, SampleSubgoalLength=~p',[StudentSubgoalLength,SampleSubgoalLength]),
	StudentSubgoalLength>SampleSubgoalLength->Errors=[idiom(error([Index], superfluous(SubgoalType,StudentSubgoalList,SampleSubgoalList)))];
	StudentSubgoalLength<SampleSubgoalLength->Errors=[idiom(error([Index], missing(SubgoalType,StudentSubgoalList,SampleSubgoalList)))],
	logging(analyse,'<compareSubgoalList2 ends > SubgoalType=~w, Errors=~p',[SubgoalType,Errors])
	.

% compareArgList(+ArgType,+StudentArgList,+SampleArgList,-Errors)
% ArgType has types: input, output, info
compareArgList(Index,ArgType,StudentArgList,SampleArgList,Errors):-
	logging(analyse,'<compareArgList1 begins> StudentArgList=~p, SampleArgList=~p',[StudentArgList,SampleArgList]),
	zuordnen(StudentArgList,SampleArgList,ids_state(10,_Pos, [], _State0),_State1 )-> Errors=[];
	        Errors=[idiom(error([Index], unmatched(ArgType,StudentArgList,SampleArgList)))],
        logging(analyse,'<compareArgList1 ends > ArgType=~w,Errors=~p',[ArgType,Errors]).

compareArgList(Index,ArgType,StudentArgList,SampleArgList,Errors):-
	logging(analyse,'<compareArgList2 begins> StudentArgList=~p, SampleArgList=~p',[StudentArgList,SampleArgList]),
	length(StudentArgList,StudentArgLength),
	length(SampleArgList,SampleArgLength),
	StudentArgLength>SampleArgLength->Errors=[idiom(error([Index], superfluous(ArgType,StudentOutputList,SampleOutputList)))];
	StudentArgLength<SampleArgLength->Errors=[idiom(error([Index], missing(ArgType,StudentOutputList,SampleOutputList)))],
	logging(analyse,'<compareArgList2 ends > ArgType=~w, Errors=~p',[ArgType,Errors])
	.

 
% find_info_list(+Head-_,+TestSubgoal,+RecCall, -InfoList)
find_info_list_for_test_process(Head,TestSubgoal,RecCall, InfoList):-
	logging(analyse,'<find_info begins>Head=~w,TestSubgoal=~w,RecCall=~w',[Head,TestSubgoal,RecCall]),
	% if the TestSubgoal does not exist, it will be an unbound variable.
	% then the Info argument is the one which occurs both in the head and in the recursive call.
	var(TestSubgoal)->
	   findall(info(InfoIndex,InfoArg),(Head=..[_|HeadArgs],
		      member(InfoArg,HeadArgs),
		      logging(analyse,'<find_info1>InfoArg=~w',[InfoArg]),
		      nth1(InfoIndex,HeadArgs,InfoArg),
		      RecCall=..[_|RecArgs],
		      member(InfoArg,RecArgs),
		      logging(analyse,'<find_info1>This InfoArg exists also in recursive subgoal',[])),
		InfoList)
;
	findall(info(InfoIndex,InfoArg),(Head=..[_|HeadArgs],
		      member(InfoArg,HeadArgs),
		      logging(analyse,'<find_info2>InfoArg=~w',[InfoArg]),
		      nth1(InfoIndex,HeadArgs,InfoArg),
		      TestSubgoal=..[_|TestArgs],
		      member(InfoArg,TestArgs),
		      logging(analyse,'<find_info2>This InfoArg exists also in test subgoal',[]), 			 
		      RecCall=..[_|RecArgs],
		      member(InfoArg,RecArgs),
		      logging(analyse,'<find_info2>This InfoArg exists also in recursive subgoal',[])),
		InfoList),
	logging(analyse,'<find_info ends>InfoList=~w',[InfoList]).

% find_input_argument_for_test_process(+Head,+TestSubgoal,+InfoList,-Input)
find_input_output_for_test_process(Clause,HeadArgs,TestArgs,[Info|InfoRest],InputList,OutputList):-
	logging(analyse,'<find_input_output1 begins>HeadArgs=~w,TestArgs=~w,Info=~w',[HeadArgs,TestArgs,Info]),
	Info=info(_,InfoArg),
	logging(analyse,'<find_input_output>InfoArg=~w',[InfoArg]),
	(   HeadArgs\=[]->select(InfoArg,HeadArgs,HeadArgsTemp);
	              HeadArgsTemp=[]),
	% select the InfoArg from TestArgs if TestArgs not empty.
	(TestArgs\=[]->select(InfoArg,TestArgs,TestArgsTemp);
	              TestArgsTemp=[]),
	find_input_output_for_test_process(Clause,HeadArgsTemp,TestArgsTemp,InfoRest,InputList,OutputList),
	logging(analyse,'<find_input_output1 ends>InputList=~w, OutputList=~w',[InputList,OutputList]).
find_input_output_for_test_process(Clause,HeadArgs,TestArgs,[],InputList,OutputList):-
	logging(analyse,'<find_input_output2 begins>HeadArgs=~w,TestArgs=~w',[HeadArgs,TestArgs]),
	findall(input(Arg,Input) , technique(Arg,list_head(Input),Clause), InputList),
	logging(analyse,'<find_input_output2>InputList=~w',[InputList]),
	findall(InputPos, 
		(member(input(InputPos,_),InputList)
		),
		InputPosList
	       ),
	logging(analyse,'<find_input_output2>InputPosList=~w',[InputPosList]),
	findall(output(Arg,Output),
		(nth1(Arg,HeadArgs,Output), not(member(Arg,InputPosList))   
		),
		OutputList
	       ),
	/**
	% select the input argument from HeadArgs then the result is the output
	TestArgs=[TestArg|TestRest],
	HeadArg=[TestArg|_],
	select(HeadArg,HeadArgs,HeadArgsTemp),
	logging(analyse,'<find_input_output2>HeadArg=~w has been selected',[HeadArg]),
	logging(analyse,'<find_input_output2>HeadArgsTemp=~w',[HeadArgsTemp]),
	find_input_output_for_test_process(HeadArgsTemp,TestRest,[],TestArgs,OutputList),
*/
	logging(analyse,'<find_input_output2 ends> InputList=~w, OutputList=~w',[TestArgs,OutputList]).
%find_input_output_for_test_process(RecCall,HeadArgs,[],[],_,HeadArgs):-
%	logging(analyse,'<find_input_output3 ends> OutputList=~w',[HeadArgs]).

% for the case that the test subgoal does not exist
%find_input_output_for_test_process(HeadArgs,[],[Info|InfoRest],InputList,OutputList):-
%	logging(analyse,'<find_input_output4 ends> OutputList=~w',[HeadArgs]).
% find_input_output_for_test_process(HeadArgs,[],[],[],HeadArgs).

% find_rest_list(+Numlist,+BaseIndexList,-InfoList)
% Numlist is a complete set of argument position [1..ListLength]
% BaseIndexList is a subset of Numlist 
find_rest_list(Numlist,[BaseIndex|Rest],InfoList):-
	logging(analyse,'<find_rest_list begins>Numlist=~w, BaseIndex=~w, Rest=~w',[Numlist,BaseIndex,Rest]),
	select(BaseIndex,Numlist,NumlistNew),
	find_rest_list(NumlistNew,Rest,InfoList),
	logging(analyse,'<find_rest_list ends>InfoList=~w',[InfoList]).
find_rest_list(Numlist,[],Numlist).

	


analyse_clause(StudentClause,SampleClause,Index,ErrorList):-
	logging(analyse,'<analyse_clause1 begins>StudentClause=~w, SampleClause=~w, Index=~w',[StudentClause,SampleClause,Index]),
	(   base_case(StudentClause,_), base_case(SampleClause,_)->
	    analyse_base_case(StudentClause,SampleClause,Index,ErrorList)
	),
	logging(analyse,'<analyse_clause1 ends>ErrorList=~w',[ErrorList]).
analyse_clause(StudentClause,SampleClause,Index,ErrorList):-
	logging(analyse,'<analyse_clause2 begins>StudentClause=~w, SampleClause=~w, Index=~w',[StudentClause,SampleClause,Index]),
	(   recursive_case(StudentClause,_,_,_), recursive_case(SampleClause,_,_,_)->
	    analyse_recursive_case(StudentClause,SampleClause,Index,ErrorList1),
	    analyse_nonrecursive_calls(StudentClause,SampleClause,Index,ErrorList2),
	    append(ErrorList1,ErrorList2,ErrorList)
	),
	logging(analyse,'<analyse_clause2 ends>ErrorList=~w',[ErrorList]).

analyse_clause(StudentClause,SampleClause,Index,ErrorList):-
	logging(analyse,'<analyse_clause4 begins>StudentClause=~w, SampleClause=~w, Index=~w',[StudentClause,SampleClause,Index]),
	(   not(base_case(StudentClause,_)), base_case(SampleClause,_)->
	   ErrorList=[idiom(error([Index], basecase_expected(SampleClause,StudentClause)))]
	),
	logging(analyse,'<analyse_clause4 ends>ErrorList=~w',[ErrorList]).
analyse_clause(StudentClause,SampleClause,Index,ErrorList):-
	logging(analyse,'<analyse_clause5 begins>StudentClause=~w, SampleClause=~w, Index=~w',[StudentClause,SampleClause,Index]),
	(   not(recursive_case(StudentClause,_,_,_)), recursive_case(SampleClause,_,_,_)->
	   ErrorList=[idiom(error([Index], recursivecase_expected(SampleClause,StudentClause)))]
	),
	logging(analyse,'<analyse_clause5 ends>ErrorList=~w',[ErrorList]).

analyse_base_case(StudentClause,SampleClause,Index,ErrorList):-
	logging(analyse,'<analyse_base_case begins>StudentClause=~w, SampleClause=~w, Index=~w',[StudentClause,SampleClause,Index]),
	base_case(StudentClause,base_constructor(IndexList1,ConstructorList1)),
	base_case(SampleClause,base_constructor(IndexList2,ConstructorList2)),
	(   IndexList1=IndexList2, ConstructorList1=ConstructorList2 ->
	         ErrorList=[];	
	         ErrorList=[idiom(error([Index],unmatched_baseconstructor(StudentClause,SampleClause)))]
	),
	logging(analyse,'<analyse_base_case ends>ErrorList=~w',[ErrorList]).

analyse_recursive_case(StudentClause,SampleClause,Index,ErrorList):-
	logging(analyse,'<analyse_recursive_case begins>StudentClause=~w, SampleClause=~w, Index=~w',[StudentClause,SampleClause,Index]),
	% look for argument technique
	findall(technique(N,Tech),technique(N,Tech,StudentClause),StudentTechList),
	findall(technique(N,Tech),technique(N,Tech,SampleClause),SampleTechList),
	findall(idiom(error([Index],Mismatch)),compare_technique(StudentTechList,SampleTechList,Mismatch),ErrorList),
	logging(analyse,'<analyse_recursive_case ends>ErrorList=~w',[ErrorList]).

analyse_nonrecursive_calls(StudentClause,SampleClause,Index,ErrorList):-
	logging(analyse,'<analyse_nonrec_case begins>StudentClause=~w, SampleClause=~w, Index=~w',[StudentClause,SampleClause,Index]),
	findall(technique(Tech),technique(Tech,StudentClause),StudentTechList),
	logging(analyse,'<analyse_nonrec_case>nonrecursive call techniques of student=~w',[StudentTechList]),
	findall(technique(Tech),technique(Tech,SampleClause),SampleTechList),
	logging(analyse,'<analyse_nonrec_case>nonrecursive call techniques of sample=~w',[SampleTechList]),
	findall(idiom(error([Index],Mismatch)),compare_technique(StudentTechList,SampleTechList,Mismatch),ErrorList),
	logging(analyse,'<analyse_nonrec_case ends>ErrorList=~w',[ErrorList]).


compare_technique(StudentTechList,SampleTechList,should_be(X)):-
	member(X,SampleTechList),
	forall(member(Y,StudentTechList), not(zuordnen(X,Y,ids_state(10,_Pos, [], _State0),_State1))),
	not(member(X,StudentTechList)).
compare_technique(StudentTechList,SampleTechList,should_not_be(X)):-
	member(X,StudentTechList),
	forall(member(Y,SampleTechList), not(zuordnen(X,Y,ids_state(10,_Pos, [], _State0),_State1))),
	not(member(X,SampleTechList)).
	
	

/**

predicate_clauses(_,[],[]).
predicate_clauses(Functor/Arity,[Head-Body|Rest],[Head-Body|Rest1]):-
	functor(Head,Functor,Arity),
	predicate_clauses(Functor/Arity,Rest,Rest1).
predicate_clauses(Functor/Arity,[Head-_|Rest],Rest1):-
	predicate_clauses(Functor/Arity,Rest,Rest1),
	not(functor(Head,Functor,Arity)).
	

%=schema(?Schema,+NKM)

% schema(basecase(X),Pred)  heisst "Pred hat X als Rekursionsabschluss".
schema(basecase(X),Pred):-
	member(X-[],Pred),
        logging(analyse,'<schema>This clause is a basecase: X=~w',[X]).
% ich frage mich, wozu diese Clause dient?
% schema(basecase,Pred)     heisst "Pred hat einen Rekursionsabschluss".
schema(basecase,Pred):-
	once(schema(basecase(_),Pred)),
	logging(analyse,'<schema>This clause is a basecase: Pred=~w',[Pred]).
% schema(recursion(X),Pred) heisst "Pred hat X als Rekursionsschritt".
schema(recursion_step(Head,Call),Pred):-
	member(Head-NBody,Pred),
	recursive_case(Head-NBody,_,Call,_),
	logging(analyse,'<schema>This clause is a recursion: Head=~w, Call=~w',[Head,Call]).
% ich frage mich, wozu diese Clause dient?
% schema(recursion,Pred)    heisst "Pred hat einen Rekursionsschritt".
schema(recursion_step,Pred):-
	once(schema(recursion_step(_,_),Pred)),
	logging(analyse,'<schema>This clause is a recursion:Pred=~w',[Pred]).
/**
Fr uns ist basecase(X) interessanter, whrend fr das Feedback
allein basecase interessant ist, wenn wir nur eine Meldung "Du brauchst
einen Rekursionsabschluss" geben wollen.
*/


/**
Ein Beispiel fr list_filter und list_recursion wre:
alle_geraden([],[]).
alle_geraden([A|L1],[A|L2]):-
	A mod 2 =:= 0,
	alle_geraden(L1,L2).
alle_geraden([A|L1],L2):-
	A mod 2 =:= 1,
	alle_geraden(L1,L2).
Das ist dann list_recursion(1), weil an der ersten Stelle immer etwas
weggenommen wird, bis die Liste leer ist, und list_filter(2), weil
an der zweiten Stelle mal etwas "weggenommen" wird und mal einfach gleich
gelassen wird.
Eigentlich wre es schner, wenn man die beiden mit Aufrufen von
technique/3 definieren wrde:

a_technique(N,Technique,Pred):-
	once((member(Clause,Pred),
		technique(N,Technique,Clause))).
every_technique(N,Technique,Pred):-
	forall((member(Clause,Pred),
		recursive_case(Clause,_,_,_)),
	       (technique(N,Technique,Clause))).

schema(list_recursion(Rec),Pred):-
	schema(basecase(Base),Pred),
	arg(Rec,Base,[]),
	a_technique(Rec,list_uncons(_),Pred),
	every_technique(Rec,list_uncons(_),Pred).
schema(list_filter(Rec),Pred):-
	schema(basecase(Base),Pred),
	arg(Rec,Base,[]),
	a_technique(Rec,list_uncons(_),Pred),
	a_technique(Rec,same(_),Pred).
*/


schema(list_recursion(RecPos),Pred):-
	logging(analyse,'<schema>test a list_recursion: Pred=~w',[Pred]),
	% if it exists a base case
	schema(basecase(Base),Pred),
        % one argument is an empty list
	arg(RecPos,Base,[]),
	% and it exists a recursion step
	once((schema(recursion_step(Head,Call),Pred),
	arg(RecPos,Head,[_|Rest]),
	arg(RecPos,Call,Rest))),
	forall(schema(recursion_step(Head1,Call1),Pred),
	       (   % the argument at the palce RecPos will be modified in a recursion step
	           arg(RecPos,Head1,[_|Rest1]),
		   arg(RecPos,Call1,Rest1))
	       ),
	logging(analyse,'<schema>there is a list_recursion: RecPos=~w',[RecPos]).
/**
list_filter(X) sollte eine Abwandlung von list_recursion(X) sein, bei dem
nicht jedesmal etwas ab- bzw. aufgebaut werden muss, sondern es auch
passieren kann, dass das Argument gleich bleiben kann.
Das Schema she dann so aus:
pred([]).
pred([A|As]):-pred(As).
pred(As):-pred(As).
*/
schema(list_filter(RecPos),Pred):-
	logging(analyse,'<schema>test a list_filter: Pred=~w',[Pred]),
	% there is a base case
	schema(basecase(Base),Pred),
	arg(RecPos,Base,[]),
	% there is a recursion step 
	once((schema(recursion_step(Head,Call),Pred),
	% so that the argument at the RecPos will be modified with this recursion step
	arg(RecPos,Head,[_|Rest]),
	arg(RecPos,Call,Rest))),
	% there is a recursion step
	once((schema(recursion_step(Head1,Call1),Pred),
	% so that the argument at the RecPos will not be modified within this recursion step.
	arg(RecPos,Head1,Rest),
	arg(RecPos,Call1,Rest))),
	logging(analyse,'<schema>There is a list_filter: RecPos=~w',[RecPos]).

/**
accu(1,3) heisst, dass auf der 1. Stelle etwas akkumuliert wird, das dann
als Ergebnis auf der 3. Stelle zurckgegeben wird.
Das Schema sieht so aus:
pred(Accu,???,Accu).
pred(Accu,???,Result):-...,pred(Accu1,???,Result).
*/

schema(accu(Accu,Return),Pred):-
	logging(analyse,'<schema>test an accu: Pred=~w',[Pred]),
	schema(recursion_step(Head0,Call0),Pred),
	arg(Return,Head0,RetVar0),
	arg(Return,Call0,RetVar0),

	schema(basecase(Base0),Pred),
	arg(Accu,Base0,AccuVar0),	
	arg(Return,Base0,AccuVar0),

	Accu\=Return,
	% verifies that in the basecase, the value of Accu and Return is the same
	forall(schema(basecase(BaseCase),Pred),
	       (arg(Accu,BaseCase,AccuVar),
		arg(Return,BaseCase,AccuVar)
	       )
	      ),
	% verifies that, in every recursion step, the value of Return is not modified.
	forall(schema(recursion_step(Head,Call),Pred),
	       (arg(Return,Head,RetVar),
		arg(Return,Call,RetVar)
	        )
	      ),
	
	logging(analyse,'<schema>There is an accu: Accu=~w; Return=~w',[Accu,Return]).


% We don't want to give these to the user.
% the base case is usually covered by testing.
uninteresting(recursion_step(_,_)).
uninteresting(basecase(_)).

compare_idioms(Idioms1,Idioms2,should_be(X)):-
	member(X,Idioms1),
	not(uninteresting(X)),
	not(member(X,Idioms2)).
compare_idioms(Idioms1,Idioms2,should_not_be(X)):-
	member(X,Idioms2),
	not(uninteresting(X)),
	not(member(X,Idioms1)).

idiom_detector(SamplePred,StudentPred,Errs):-
	logging(analyse,'<idiom_detector>Input Sample Solution: SamplePred=~w',[SamplePred]),
	logging(analyse,'<idiom_detector>Input Student Solution: StudentPred=~w',[StudentPred]),
	% just test this predicate
	analyse_predicates(StudentPred,SamplePred,Fehler),
	logging(analyse,'<idiom_detector>Fehler beim Matchen: ~w',[Fehler]),
	findall(Idiom,schema(Idiom,SamplePred),Idioms1),
	logging(analyse,'<idiom_detector>Idioms1: =~w',[Idioms1]),
	findall(technique(N,Tech),(member(Clause,SamplePred),technique(N,Tech,Clause)),TechList1),
	logging(analyse,'<idiom_detector>TechList1: =~w',[TechList1]),
	findall(Idiom,schema(Idiom,StudentPred),Idioms2),
	logging(analyse,'<idiom_detector>Idioms2: =~w',[Idioms2]),
	findall(technique(N,Tech),(member(Clause,StudentPred),technique(N,Tech,Clause)),TechList2),
	logging(analyse,'<idiom_detector>TechList2: =~w',[TechList2]),
	findall(Mismatch,compare_idioms(Idioms1,Idioms2,Mismatch),Errs),
	logging(analyse,'<idiom_detector>Errs=~w',[Errs]).

*/
%testcase(rev,rev([1,2,3],[],_X)).
%musterlsg(rev,[rev([],X,X)-[],
%	       rev([A|X],Y,Z)-[rev(X,[A|Y],Z)]
%	      ]):-
%	replaceVars(['A'=A,'X'=X,'Y'=Y,'Z'=Z]).

/**
The craft of Prolog, p.22
The input argument corresponding to the initial state of a conventional
variable is called an accumulator. THe output argument corresponding to
the final state of a conventional variable is called a result. THe two
arguments together are called an accumulator pair.
for example: 
rev([],Reverse,Reverse).
rev([Head|Tail],Reverse0,Reverse):-
	rev(Tail,[Head|Reverse0],Reverse).
*/
/**
Typically, anb accumulator pair shows up as
p(...,Foo0, Foo,...):-
	...,
	q(...,Foo0,Foo1,...),
	...,
	r(Foo1,Foo,..),
	...
where each of the "Foo*" variables appears twice in an execution path.
*/
% arg_type(+Pred,-Arg,-accumulator(X))

/**
arg_type(Pred,Arg,accumulator(Value)):-
	basecase(Pred,BaseCase),
	recursive_case(Clause,_,_,_),
	arg(Arg,BaseCase,Value),
	variable(Pred,Arg),
	technique(Arg,list_goal([_X|_Xs]),Clause).
	
arg_type(Pred,Arg,recursive(Xs)):-
	recursive_case(Clause,_,_,_),
	technique(Arg,list_goal([_X|Xs]),Clause).
	
*/

/**
an argument is a variable if its value in every clause is variable.
*/
/**
variable(Pred,Arg):-
	forall((member(Clause,Pred),
	       arg(Arg,Clause,Value)
	       ),
	       var(Value)
	      ).	
*/


	  




