:- module(normalform,[consumeFile/2, consumeFiles/2,
		      eatStream/2,
		      consumeAtom1/2, consumeAtom/2,
		      eatFile/2,
		      replaceVars/1, normalize/2,
		      substVars/3,
		      getPredicate/4,getPredicates/2,
		      putVariables/1,
		      unfreeze/2, unfreeze/3,
		      write_the_term/1, write_the_term/4, 
		      eliminateVarIndInTerm/2]).
% changed: $Id: normalform.pl,v 1.12.2.1 2004/04/14 10:18:51 nthinh Exp $

:- use_module(logging).

%=consumeFile(+Filename, -Liste)
% liest eine Datei mit Prolog-Klauseln
% und gibt die Klauseln in Liste zurck.
% Variablen werden dabei durch '<var>'(Variablenname) ersetzt.
consumeFile(Filename,Liste):-
	open(Filename,read,Fd),
	consumeStream(Fd,Liste,[]),
	close(Fd).

%=consumeFiles(+Filenames,-Liste)
% liest mehrere Dateien mit Prolog-Klauseln
% und verarbeitet sie dann wie consumeFile
consumeFiles([Filename|Rest],Liste):-
	open(Filename,read,Fd),
	consumeStream(Fd,Liste,LR),
	close(Fd),
	consumeFiles(Rest,LR).
consumeFiles([],[]).

%=consumeAtom(+Atom,-Klausel)
% liest einen Term aus einem Atom und wandelt die Variablen
% in Strukturen um.
consumeAtom1(Atom,Klausel):-
	catch(atom_to_term(Atom,Klausel,Bindings),Error,
	      (Klausel=erronous_term,
	       Bindings=[],
	       assert(user:current_warning(Error)))),
	replaceVars(Bindings),
	free_variables(Klausel,AnonymousVars),
	replaceAnonymous(AnonymousVars).

%=eliminateVarIndInTerm(+Subgoals,-SubgoalsNoVarInd)
% look for variable indicators <var> or <a-var> and eliminate it 
eliminateVarIndInTerm(Klausel, KlauselNoVarInd):-
        Klausel =.. ArgList,
	findall(NewArg, (member(Arg, ArgList), replaceVarInd(Arg,NewArg)),NewArgList),
	KlauselNoVarInd =.. NewArgList.
replaceVarInd(Var, NewVar):-
	Var='<var>'(Value)->NewVar=Value ;
	     (	 Var='<a-var>'(Value)-> NewVar=Value;
	                                NewVar=Var
	     ).



%=consumeAtom(-Atom,+Liste)
% Liest mehrere durch '.' getrennte Klauseln aus einem Atom
consumeAtom(Atom,Liste):-
	concat_atom(L1,'.',Atom),
	once(append(L2,[''],L1);
	     (append(L2,[Last],L1),
	      assert(user:current_warning(stuff_ignored(Last))))),
	maplist(consumeAtom1,L2,Liste).

consumeStream(Fd,XS,XR):-
	read_term(Fd,X,[variable_names(Vnames),variables(_)]),
	replaceVars(Vnames),
	(  X=end_of_file ->
		XS=XR ;
		XS=[X|Rest],consumeStream(Fd,Rest,XR)
	).

replaceVars([]).
replaceVars(['_'='<a-var>'(Name)|Rest]):-!,
	gensym('_',Name),
	replaceVars(Rest).
replaceVars([Name='<var>'(Name)|Rest]):-replaceVars(Rest).

replaceAnonymous([]).
replaceAnonymous(['<a-var>'(Name)|Rest]):-
	gensym('_',Name),
	replaceAnonymous(Rest).

%=putVariables(?Term)
% ersetzt freie Variablen in <i>Term</i> durch
% entsprechende Strukturen
putVariables(Term):-
	free_variables(Term,VarList),
	putVars(VarList,1).
putVars([],_).
putVars(['<var>'(VarName)|Rest],N):-
	atom_concat('Var',N,VarName),
	succ(N,N1),
	putVars(Rest,N1).


substVars(Atom,_,_):-
	atomic(Atom),!.
substVars(Var,Varnames,_):-
	var(Var),!,
	(   
	member(Name=Var0,Varnames),Var==Var0->
	Var='<var>'(Name);
	gensym('_',Gensym),
	    Var='<a-var>',Gensym).
substVars(_,_,0):-!.
substVars(Term,Varnames,Depth):-
	succ(Depth0,Depth),
	Term=..[_|Args],
	substVars1(Args,Varnames,Depth0).
substVars1([],_,_).
substVars1([Term|Rest],Varnames,Depth):-
	substVars(Term,Varnames,Depth),
	substVars1(Rest,Varnames,Depth).

%=normalize(+Klausel,-NormalisierteKlausel)
% trennt Klauselkopf und Klauselkrper und
% wandelt den Klauselkrper in eine Liste um.
normalize((Head:-Body),Head-BodyList):- !,
	normalizeBody(Body,BodyList,[]).
normalize(Head,Head-[]).


%=normalizBody(+ClauseBody,-BodyList,[])
% change the ClauseBody in this form: (append(Rest1,[Element|Rest2],List),length(Rest1,Pos),append(Rest1,Rest2,Rest))
% to BodyList of this form: [append(_G989, [_G986|_G987], _G991), length(_G989, _G994), append(_G989, _G987, _G998)] 
normalizeBody((A,B))-->
	{!},
	normalizeBody(A),
	normalizeBody(B).
normalizeBody(X)-->[X].

%= getPredicate(Functor/Arity, ListIn, ListPred, ListNonPred)
% sucht die Klauseln zu einem Prdikat heraus
getPredicate(_,[],[],[]).
getPredicate(Functor/Arity,[Clause|Rest],[Clause|LPred],LNonPred):-
	Clause=Head-_,
	functor(Head,Functor,Arity),!,
	getPredicate(Functor/Arity,Rest,LPred,LNonPred).
getPredicate(Functor/Arity,[Clause|Rest],LPred,[Clause|LNonPred]):-
	getPredicate(Functor/Arity,Rest,LPred,LNonPred).


getPredicates([],[]).
getPredicates([Clause|Rest],[[Clause|LPred]|Others]):-
	Clause=Head-_,
	functor(Head,Functor,Arity),
	getPredicate(Functor/Arity,Rest,LPred,LNonPred),
	getPredicates(LNonPred,Others).

eatStream(Stream,PredList):-
	consumeStream(Stream,Unnormalized,[]),
	maplist(normalize,Unnormalized,Normalized),
	getPredicates(Normalized,PredList).

eatFile(FileName,PredList):-
	consumeFile(FileName,Unnormalized),
	maplist(normalize,Unnormalized,Normalized),
	getPredicates(Normalized,PredList).


% unfreeze(+Term,-Unfreezed,-Vars)
% makes a normal term out of one with <var> and <a-var> tags
% Vars contains the variables corresponding to the variable names
unfreeze(Term,Unfreezed,Vars):-
	unfreeze(Term,Unfreezed,[],Vars).
unfreeze(Term,Unfreezed):-
	unfreeze(Term,Unfreezed,[],_).
unfreeze('<var>'(Name),Var,VarsA,VarsB):-!,(
	member(Name=Var,VarsA)->VarsB=VarsA;
	VarsB=[Name=Var|VarsA]).
unfreeze('<a-var>'(_),_,Vars,Vars):-!.
unfreeze(Term0,Term1,VarsA,VarsB):-
	Term0=..[Functor|Rest0],
	unfreeze_l(Rest0,Rest1,VarsA,VarsB),
	Term1=..[Functor|Rest1].
unfreeze_l([],[],Vars,Vars).
unfreeze_l([Term0|Terms0],[Term1|Terms1],VarsA,VarsC):-
	unfreeze(Term0,Term1,VarsA,VarsB),
	unfreeze_l(Terms0,Terms1,VarsB,VarsC).

% write_the_term(Term)
% writes HTML code for the term
write_the_term(Term):-
	write_the_term(Term,10,2000,yes).

write_the_term(Var,_,_,_):-var(Var),!,
	gensym('_',Gensym),
	Var='<a-var>'(Gensym),
	sub_atom(Gensym,1,_,0,NN),
	writef('<font color="blue">_<sub>%w</sub></font>',[NN]).	

write_the_term('<var>'(Name),_,_,_):- !,
	xml_quote_attribute(Name,NameQ),
	writef('<font color="blue">%w</font>',[NameQ]).
write_the_term('<a-var>'(N),_,_,_):-!,
	sub_atom(N,1,_,0,NN),
	writef('<font color="blue">_<sub>%w</sub></font>',[NN]).

write_the_term(Term,_,_,_):-
	atomic(Term),!,
	xml_quote_attribute(Term,TermQ),
	write(TermQ).

write_the_term(_,0,_,_):-!,
	write('...').

write_the_term((A,B),Depth,_,no):-!,
	succ(Depth1,Depth),
	write('('),
	write_the_term((A,B),Depth1,2000,yes),
	write(')').

write_the_term([A|B],Depth,_,_):-
	succ(Depth1,Depth),
	write('['),
	write_the_term(A,Depth1,2000,no),
	write_list(B,Depth1).

write_the_term(Term,Depth,MaxPrec,Comma):-
	Term=..[Functor|Rest],
	succ(Depth1,Depth),
	(current_op(Prec,Type,Functor)*->
	(
	Type=xfx,Rest=[L,R]->
	succ(Prec1,Prec),
	 write_binary(Depth1,MaxPrec,Prec,Functor,L,Prec1,R,Prec1,Comma);
	Type=yfx,Rest=[L,R]->
	succ(Prec1,Prec),
	 write_binary(Depth1,MaxPrec,Prec,Functor,L,Prec,R,Prec1,Comma);
	Type=xfy,Rest=[L,R]->
	succ(Prec1,Prec),
	 write_binary(Depth1,MaxPrec,Prec,Functor,L,Prec1,R,Prec,Comma);
	Type=fx,Rest=[X]->
	succ(Prec1,Prec),
	 write_unary(Depth1,MaxPrec,Prec,Functor,X,Prec1,Comma));
	xml_quote_attribute(Functor,FunctorQ),
	 write(FunctorQ),
	 write('('),
	 write_arglist(Rest,Depth1),
	 write(')')
	).

	
write_binary(D,MP,P,F,A1,P1,A2,P2,C):-
	P>=MP->
	write('('),
	write_binary(D,2000,P,F,A1,P1,A2,P2,yes),
	write(')');
	write_the_term(A1,D,P1,C),
	xml_quote_attribute(F,Fq),
	write(' '),
	write(Fq),
	write(' '),
	write_the_term(A2,D,P2,C).
write_unary(D,MP,P,F,A1,P1,C):-
	P>=MP->
	write('('),
	write_unary(D,2000,P,F,A1,P1,yes),
	write(')');
	xml_quote_attribute(F,Fq),
	write(Fq),
	write_the_term(A1,D,P1,C).

write_arglist([A],Depth):-!,
	write_the_term(A,Depth,2000,no).
write_arglist([A|B],Depth):-
	write_the_term(A,Depth,2000,no),
	write(','),
	write_arglist(B,Depth).

write_list([],_):-!,write(']').
write_list([A|B],Depth):-!,
	write(','),
	write_the_term(A,Depth,2000,no),
	write_list(B,Depth).
write_list(X,Depth):-
	write('|'),
	write_the_term(X,Depth,2000,yes),
	write(']').
