:- module(plht,[process_plht/2,compile_plht/3,coerce_to_atom/2]).

:- use_module(library('http/http_client')).
:- use_module(library(sgml)).
:- use_module(session).
:- use_module(logging).

:- dynamic special_tag/2.
:- dynamic init_tag/2.
:- dynamic plht_init/1.


process_plht(Real_Path,Query):-
	hash_term(Real_Path,Hash),
	atom_concat('plht_',Hash,PredName),
	atom_concat(PredName,'.pl',FileName),
	once((
	     current_predicate(PredName/1),
	      time_file(FileName,TimePL),
	      time_file(Real_Path,TimePLHT),
	      TimePLHT<TimePL ;
	     logging(plht,'Compiling ~w to ~w',[Real_Path,FileName]),
	      compile_plht(PredName,Real_Path,FileName),
	      logging(plht,'Compile done',[]),
	      consult(FileName))),
	set_session_from_request(Query),
	format('Content-type: text/html\r\n'),
	write_session_cookie,
	format('\r\n'),
	call(PredName,Query),
	retract_session.


process_include(Name,Attrs,FormForContent):-
	concat_atom(['plht_inc/',Name,'.inc'],SourceFile),
	atom_concat('plht_inc_',Name,PredName),
	atom_concat(PredName,'.pl',FileName),
	once((
	     current_predicate(PredName/1),
	      time_file(FileName,TimePL),
	      time_file(SourceFile,TimeInc),
	      TimeInc<TimePL ;
	     logging(plht,'Compiling ~w to ~w',[SourceFile,FileName]),
	      compile_inc(PredName,SourceFile,FileName),
	      logging(plht,'Compile done.',[]),
	      consult(FileName))),
	call(PredName,Attrs,FormForContent).

% sort the key/value lists so that
% vars with the same name get unified.
merge_vars(A,B,Vars):-
	append(A,B,AB),
	list_to_set(AB,Vars).

merge_vars(A,B,C,Vars):-
	merge_vars(A,B,AB),
	merge_vars(AB,C,Vars).
merge_vars(A,B,C,D,Vars):-
	merge_vars(A,B,AB),
	merge_vars(C,D,CD),
	merge_vars(AB,CD,Vars).

coerce_to_atom(X,X):-atomic(X),X\=[],!.
coerce_to_atom(A-B,X):-
	coerce_to_atom(A,X1),
	coerce_to_atom(B,X2),
	concat_atom([X1,X2],'-',X),!.
coerce_to_atom([X],X).
coerce_to_atom([X|Rest],A):-
	coerce_to_atom(Rest,B),
	concat_atom(X,B,A),!.
coerce_to_atom(_,'++not-convertible++').


special_tag('pl:loop',pl_loop).
special_tag('pl:parameter',pl_parameter).
special_tag('pl:cond',pl_cond).
special_tag('pl:localvar',pl_localvar).
special_tag('pl:sessionparam',pl_sessionparam).
special_tag('pl:include',pl_include).
special_tag('pl:content',pl_content).

init_tag('pl:require',pl_init_require).
init_tag('pl:predicate',pl_init_taglet).

pl_init_require([file=Filename],[])-->[:-ensure_loaded(Filename)].
pl_init_require([module=Modulename],[])-->[:-use_module(Modulename)].

pl_init_taglet([head=Head],Content)-->
	{ 
	  form_from_tree(Content,Vars,FormForContent),
	  atom_to_term(Head,HeadTerm,Bindings),
	  merge_vars(Vars,Bindings,_)
	},
	[ HeadTerm :- FormForContent ].

pl_loop(Attrs,Content,Vars,Form):-
	memberchk(query=QueryA,Attrs),
	atom_to_term(QueryA,Term,Vars0),
	sort(Vars0,Vars1),
	form_from_tree(Content,Vars2,FormForContent),
	(   memberchk(counter=VarName,Attrs)->
	merge_vars([VarName=Var|Vars1],Vars2,Vars),
	    gensym(loop_counter,LID),
	    Form=(flag(LID,_,1),ignore((Term,
				      flag(LID,Var,Var+1),
				      once(FormForContent),fail)));
	merge_vars(Vars1,Vars2,Vars),
	    Form=ignore((Term,once(FormForContent),fail))).

pl_parameter(Attrs,Content,Vars,Form):-
	memberchk(var=VarName,Attrs),
	memberchk(name=ParamName,Attrs),
	form_from_tree(Content,Vars2,FormForContent),
	(   memberchk(list=true,Attrs) ->
	Form=(findall(X,member(ParamName=X,Search),Var),
	      FormForContent)
	;   
	once(memberchk(default=Default,Attrs);Default=''),
	    Form=(once(memberchk(ParamName=Var,Search);Var=Default),
		  FormForContent)),
	sort(['Search'=Search,VarName=Var],Vars1),
	merge_vars(Vars1,Vars2,Vars).

% versions:
% var=Var, name=Name, default=Default
%   set session parameter from parameter and unify with var,
%   defaulting to Default
% name=Name
%   set session parameter from parameter or leave it as is
pl_sessionparam(Attrs,Content,Vars,Form):-
	flag(plht_compile_mode,plht,plht),
	memberchk(var=VarName,Attrs)->
	memberchk(name=ParamName,Attrs),
	form_from_tree(Content,Vars2,FormForContent),
	once(memberchk(default=Default,Attrs);Default=''),
	Form=((memberchk(ParamName=Var,Search)->
	          session_setValue(ParamName,Var);
	      session_getValue(ParamName,Var)->true;
	      Var=Default),FormForContent),
	sort(['Search'=Search,VarName=Var],Vars1),
	merge_vars(Vars1,Vars2,Vars);
	memberchk(name=ParamName,Attrs),
	(memberchk(default=Default,Attrs)->
	Form0=session_setValue(ParamName,Param);
	Form0=true),
	form_from_tree(Content,Vars2,FormForContent),
	Form=((memberchk(ParamName=Param,Search)->
	        session_setValue(ParamName,Param);
	      session_getValue(ParamName,Param)->true;
	      Form0),
	      FormForContent),
	merge_vars(['Search'=Search],Vars2,Vars).
	

% pl_localvar var=Varname
% make Varname a variable local to the extent of this tag
pl_localvar([var=Varname],Content,Vars,Form):-
	form_from_tree(Content,Vars2,Form),
	once(delete(Varname=_,Vars2,Vars);Vars=Vars2).

% pl_cond
% pl_if test=Test
% pl_else
% conditionally execute parts of the template
pl_cond(_,[],[],true).
pl_cond(Attrs,[element('pl:if',[test=Test],Content)|Rest],Vars,Form):-
	!,atom_to_term(Test,TermT,Vars1),
	form_from_tree(Content,Vars2,FormForContent),
	pl_cond(Attrs,Rest,Vars3,FormForRest),
	Form=(TermT->FormForContent;FormForRest),
	merge_vars(Vars1,Vars2,Vars3,Vars).
pl_cond(_,[element('pl:else',_,Content)|_],Vars,Form):-
	!,
	form_from_tree(Content,Vars,Form).
pl_cond(Attrs,['\n'|Rest],Vars,Form):-
	!,pl_cond(Attrs,Rest,Vars,Form).
pl_cond(Attrs,[Something|Rest],Vars,Form):-
	form_from_tree1(Something,Vars1,FormS),
	pl_cond(Attrs,Rest,Vars2,FormForRest),
	Form=(FormS,fail;FormForRest),
	merge_vars(Vars1,Vars2,Vars).

% pl_include name=IncludeName, ...		      
% includes a .inc file
% The attribute values (text) are accessible to the include
% via pl:parameter tags
%!!! We should make it possible to use variables as parameters
%!!! by setting an attribute to $Var
pl_include(Attrs,Content,Vars,Form):-
	form_from_tree(Content,Vars1,FormForContent),
	select(name=Name,Attrs,Attrs2),!,
	attrs_to_values(Attrs2,Attrs3,Vars2),
	merge_vars(Vars1,Vars2,Vars),
	Form=process_include(Name,Attrs3,FormForContent).

%=attrs_to_values(+AttrsIn,-AttrsOut,-Vars)
% looks for $Varname constructs in AttrsIn
attrs_to_values([],[],[]).
attrs_to_values([Name=DVar|Rest1],
		[Name=Var|Rest2],
		Varnames):-
	atom_concat('$',Varname,DVar),!,
	attrs_to_values(Rest1,Rest2,Varnames0),
	merge_vars(Varnames0,[Varname=Var],Varnames).
attrs_to_values([Name=Value|Rest1],
		[Name=Value|Rest2],
		Varnames):-
	attrs_to_values(Rest1,Rest2,Varnames).

% includes the content of the pl:include tags.
% this can be used to wrap contents between pl:include tags
pl_content([],_,['$Content'=Content],call(Content)):-
	flag(plht_compile_mode,inc,inc).

%=form_for_value(ValA,Vars,Form)
% generates the form for printing out an attribute for a
% html tag
% values of the form $Varname
% or of the form !!URL
% are handled specially
form_for_value(ValA,Vars,Form):-
	atom_concat('$',VarName,ValA),!,
	Vars=[VarName=X],Form=(coerce_to_atom(X,X1),
			       xml_quote_attribute(X1,Y),write(Y)).
	
form_for_value(ValA,Vars,Form):-
	atom_concat('!!',URL,ValA),!,
	atom_codes(URL,Codes),
	phrase(url:uri(http,Parts),Codes),
	xform_parts(Parts,Parts1,Vars),
	Form=(encode_url2(Parts1,X1),write(X1)).

form_for_value(ValA,[],Form):-
	xml_quote_attribute(ValA,Val),Form=write(Val).

xform_parts(Parts,[search(Search1)|Parts0],Vars):-
	select(search(Search),Parts,Parts0),!,
	xform_search(Search,Search1,Vars0),
	sort(Vars0,Vars).
xform_parts(Parts,Parts,[]).

xform_search([],[],[]).
xform_search([Key=DVarname|RestIn],
	     [Key=Var|RestOut],
	     [Varname=Var|RestVars]):-
	atom_concat('$',Varname,DVarname),!,
	xform_search(RestIn,RestOut,RestVars).
xform_search([X|RestIn],[X|RestOut],Vars):-
	xform_search(RestIn,RestOut,Vars).
	      


form_from_attrs([],[],true).
form_from_attrs([Name=Value|Rest],Vars,Form):-
	form_for_value(Value,Vars1,FormForValue),
	form_from_attrs(Rest,Vars2,FormForRest),
	Form=(write(' '),write(Name),write('="'),FormForValue,write('"'),FormForRest),
	merge_vars(Vars1,Vars2,Vars).

form_for_tag(Tag,Attrs,Content,Vars,Form):-
	% special tag
	special_tag(Tag,Pred)->
	call(Pred,Attrs,Content,Vars,Form);
	% init tag
	init_tag(Tag,Pred)->
	retract(plht_init(X)),
	(call(Pred,Attrs,Content,X1,X)->
	assert(plht_init(X1)),
	 Form=true;
	assert(plht_init(X)),Form=write('<!-- inittag not compiled -->'));
	% html tag
	form_from_tree(Content,Vars1,FormForContent),
	form_from_attrs(Attrs,Vars2,FormForAttrs),
	merge_vars(Vars1,Vars2,Vars),
	Form=(write('<'),write(Tag),FormForAttrs,write('>'),
	      FormForContent,write('</'),write(Tag),write('>')).

form_from_tree1(element(Tag,Attrs,Content),Vars,Form):-
	form_for_tag(Tag,Attrs,Content,Vars,Form).

form_from_tree1(CDATA,[],Form):-
	atom(CDATA),
	xml_quote_cdata(CDATA,Quoted),
	Form=write(Quoted).

form_from_tree1(pi(F0),Vars,Term):-
	(atom_concat('pl ',F1,F0);atom_concat('pl\n',F1,F0))->
	atom_to_term(F1,Term,Vars);
	atom_concat('pl~',F1,F0)->
	atom_to_term(F1,Term0,Vars),
	    Term=(xml_quote_attribute(Term0,Quoted),
		  write(Quoted));
	atom_concat('pl=',F1,F0)->
	atom_to_term(F1,Term0,Vars),
	    Term=(term_to_atom(Term0,Term1),
		  xml_quote_attribute(Term1,Quoted),
		  write(Quoted));
	atom_concat('verbatim ',F1,F0)->
	Term=write(F1),Vars=[].

form_from_tree([Something|Rest],Vars,Form):-
	form_from_tree1(Something,Vars1,Form1),
	form_from_tree(Rest,Vars2,Form2),
	merge_vars(Vars1,Vars2,Vars),
	Form=(Form1,Form2).
form_from_tree([],[],true).


form_from_file(File,Vars,Form):-
	load_xml_file(File,Tree),
	form_from_tree(Tree,Vars,Form0),
	optimize(Form0,Form).

compile_plht(PredName,InFile,OutFile):-
	flag(plht_compile_mode,_,plht),
	assert(plht_init([])),
	form_from_file(InFile,Vars,Form),
	retract(plht_init(Inits)),
	merge_vars(Vars,['Query'=Query,'Search'=Search],_),
	Pred=..[PredName,Query],
	tell(OutFile),
	write_ln('% This file is autogenerated by plht.'),
	forall(member(Init,Inits),
		portray_clause(Init)),
	writeq(Pred:-
	       ((member(method(post),Query)->http_read_data(Query,Data,[]);
		Data=[]),
		once(memberchk(search(Search0),Query);Search0=[]),
		append(Data,Search0,Search),
		Form)),
	write('.'),
	told.

compile_inc(PredName,InFile,OutFile):-
	flag(plht_compile_mode,_,inc),
	assert(plht_init([])),
	form_from_file(InFile,Vars,Form),
	retract(plht_init(Inits)),
	merge_vars(Vars,['Search'=Attrs,'$Content'=Content],_),
	Pred=..[PredName,Attrs,Content],
	tell(OutFile),
	write_ln('% This file is autogenerated by plht.'),
	forall(member(Init,Inits),
	       portray_clause(Init)),
	writeq(Pred:-Form),
	write('.'),
	told.

linearize((A,B))-->!,linearize(A),linearize(B).
linearize(A1)-->[A],{optimize(A1,A)}.

unlinearize([A],A).
unlinearize([A|As],(A,AA)):-
	unlinearize(As,AA).

optimize1([A1,true|As],Bs):-!,
	optimize1([A1|As],Bs).
optimize1([write(T1),write(T2)|Rest],Bs):-
	atom(T1),atom(T2),!,
	atom_concat(T1,T2,T3),
	optimize1([write(T3)|Rest],Bs).
optimize1([A|Rest],[A|Bs]):-
	optimize1(Rest,Bs).
optimize1([],[]).
optimize((A,A1),B):-linearize((A,A1),As,[]),optimize1(As,Bs),
	unlinearize(Bs,B),!.
optimize(ignore(A),ignore(A1)):-optimize(A,A1).
optimize(once(A),once(A1)):-optimize(A,A1).
optimize(A,A).


