% Author: Yannick Versley, Nguyen-Thinh Le
% Modified by: $Id: syntaxChecker.pl,v 1.10 2004/04/02 13:15:31 nthinh Exp $

/**
 This module provides the methods to check the syntax of a prolog clause or query.
*/

:-module(syntaxChecker,[check_clause_syntax/3,
			check_query_syntax/3,
			check_program_syntax/3,
		        check_program_syntax2/3]).
:-use_module(ids).
:-use_module(struct).
:-use_module(listUtil).
:- use_module(logging).

:-declare_struct(parser_state,[string]).

% check_clause_syntax(+String,-Term,-Errors)
% check_query_syntax(+String,-Term,-Errors)
% reads a string of Prolog clause, returns this clause in term and shows the positions of syntax errors

check_program_syntax2(String,Clauses,Errors):-
	ids_search(do_program(Clauses),[0],
		   parser_state(String),_,
		   0,50,
		   Errors).

% delete([X|Rest],X,Rest).
% delete([Y|Rest],X,[Y|Result]):-delete(Rest,X,Result).
% test program: [100, 101, 108, 101, 116, 101, 40, 91, 88, 124, 82, 101, 115, 116, 93, 44, 88, 44, 82, 101, 115, 116, 41, 46, 10, 100, 101, 108, 101, 116, 101, 40, 91, 89, 124, 82, 101, 115, 116, 93, 44, 88, 44, 91, 89, 124, 82, 101, 115, 117, 108, 116, 93, 41, 58, 45, 100, 101, 108, 101, 116, 101, 40, 82, 101, 115, 116, 44, 88, 44, 82, 101, 115, 117, 108, 116, 41, 46, 10]
% this is an alternative solution for recurson exercise
check_program_syntax(String,Clauses,Errors):-
	logging(syntaxChecker,'<check_program_syntax>Input String=~w',[String]),
	seperateListByX(String,10,ClauseList),     % seperate a program to a list of clauses
	logging(syntaxChecker,'<check_program_syntax>ClauseList=~w',[ClauseList]),
	findall([Term,N-Errors],(member(Clause, ClauseList),check_clause_syntax(Clause,Term, Errors),nth1(N,ClauseList,Clause)),TermErrorsList),
	logging(syntaxChecker,'<check_program_syntax>TermErrorsList=~w',[TermErrorsList]),
	seperateTerm_Errorlist(TermErrorsList,Clauses,Errors).


check_clause_syntax(String,Term,Errors):-
	% if the last element is a space then remove spaces
	remove_first_space_in_list(String,NewString1),
	remove_last_space_in_list(NewString1,NewString2),
	ids_search(do_clause(Term),parser_state(NewString2),Errors).

% this is for database query exercise
check_query_syntax(String,Term,Errors):-
	remove_first_space_in_list(String,NewString1),
	remove_last_space_in_list(NewString1,NewString2),
	ids_search(do_query(Term),parser_state(NewString2),Errors).


% just for testing purposes
check_list_syntax(String,Term,Errors):-
	ids_search(do_list(Term),[],parser_state(String),_NewState,0,50,Errors).
check_terms_syntax(String,Term,Errors):-
	ids_search(do_terms(Term),[0],parser_state(String),_NewState,0,50,Errors).
do_list(List) --> list(List),end.

do_terms(Terms) --> terms(Terms),end.

% definition of clause
% test: syntaxChecker:sentence(Sen,ids_state(10,pos,[],parser_state("a(B):-c(D,E)e(_)")), ids_state(Points,pos,Errors,Rest)).
do_clause(Term)-->sentence(Term),end.
% definition of a query
do_query(Query) --> body(Query),end.




do_program([])-->end.
do_program([Term|Rest])-->
	position([N|Pos],[N1|Pos]),
	{succ(N,N1)},
	opt_white,
	sentence(Term),
	{logging(syntaxChecker,'<do_program>Term=~w',[Term])},
	do_program(Rest),
	opt_white,opt_end_of_line.

%do_program(Program)-->
%{logging(syntaxChecker,'<do_program>Program=~w',[Program])},
%program(Program),end.


%------ Prolog syntax rules definition ------------%
program(Sentence) --> sentence(Sentence).
program([Sentence|Program]) --> sentence(Sentence),line_end,program(Program).

sentence(Head-Body) --> head(Head), eat(":-"), opt_white,body(Body), one_time(eat("."),'.'),opt_white,opt_end_of_line,opt_end_of_line.
sentence(Head-[])-->head(Head),one_time(eat("."),'.'),opt_white,opt_end_of_line,opt_end_of_line.

head(Head) --> with_position(0,goal(Head)).

body([])-->error(missing_body,15).
body(Body)-->with_position(1,with_position(1,goals(Body))).

goal(Term) --> compound_term(Term).
goal(X is Expr)-->term(X), whitespace, eat("is"),whitespace,expr(Expr).

goal(Term)-->term(X),opt_white,infix_operator2(Op),opt_white,term(Y),{Term=..[Op,X,Y]}.

goal(Term)-->expr(X),opt_white,infix_operator3(Op),opt_white,term(Y), {Term=..[Op,X,Y]}.

goals([Goal])-->goal(Goal).

goals([Goal|Goals]) --> goal(Goal), comma, 
	position([N|Pos],[N1|Pos]),{succ(N,N1)},
	goals(Goals).

comma-->one_time(eat(","),','),opt_white.

opt_white-->[].
opt_white-->whitespace.

whitespace-->eat([X]),{code_type(X,white)}.

new_line--> eat([X]), {code_type(X,end_of_line), X=10}.
cr--> eat([X]), {code_type(X,end_of_line), X=13}.
opt_end_of_line-->[].
opt_end_of_line-->new_line.
opt_end_of_line-->cr.


compound_term(Term) --> atom(F), left_parenthesis, with_position(1,terms(Args)), right_parenthesis,
	{Term=..[F|Args]}.
compound_term(Term) --> list(Term).

compound_terms([T]) --> compound_term(T).
compound_terms([T|Ts]) --> compound_term(T), comma, compound_terms(Ts).


expr(Expr)-->simpleExpr(Expr).
expr(Term)-->simpleExpr(Expr1),infix_operator(Op),simpleExpr(Expr2),{Term=..[Op,Expr1,Expr2]}.

simpleExpr(Expr)-->term(Expr).
simpleExpr(Expr)-->left_parenthesis,expr(Expr),right_parenthesis.

left_parenthesis --> one_time(eat("("),'(').
right_parenthesis --> one_time(eat(")"),')').

term(T) --> constant(T).
term('<var>'(Varname)) --> variable(Varname).
term('<a-var>'(Gensym))--> eat("_"), {gensym('_',Gensym)}.
term(T) --> compound_term(T).


terms([T]) --> term(T).
terms([T|Ts]) --> term(T), comma,
	position([N|Pos],[N1|Pos]),{succ(N,N1)},
	terms(Ts).
% case: a()
terms([])-->error(forgotten_argument,5).



prefix_operation --> prefix_operator, goal.

postfix_operation --> goal, postfix_operator.

infix_operation([Term1,Term2]) --> goal(Term1), infix_operator, goal(Term2).

% ??? need details
prefix_operator.

% ??? need details
postfix_operator.

% ??? need extension
infix_operator(+) --> one_time(eat("+"),'+'). % "['+'].
infix_operator(-) --> one_time(eat("-"),'-').
infix_operator(*) --> one_time(eat("*"),'+').
infix_operator(/) --> one_time(eat("/"),'/').
infix_operator(mod) --> eat("mod").
%infix_operator([]) --> error(missing_operator,15).
%infix_operator(is) --> eat("is").
infix_operator2(==) --> eat("==").
infix_operator2(=) --> eat("=").
infix_operator2(\=)--> eat("\\=").
infix_operator2(=\=) --> eat("=\\=").
infix_operator3(=<) --> eat("=<").
infix_operator3(<) --> eat("<").
infix_operator3(>) --> eat(">").
infix_operator3(>=) --> eat(">=").
infix_operator3(@=) --> eat("@=").
infix_operator3(=:=) --> eat("=:=").


list([]) --> eat("[]").
list(Terms) --> left_bracket, terms(Terms),right_bracket.
list([ListHeads|Term]) --> left_bracket, list_heads(ListHeads), one_time(eat("|"), '|'), term(Term), right_bracket.

% because definition of terms returns terms in this form [a,b,x] which will produce the following list if we apply it for the 3rd list definition: [[a,b,X]|Y].
% But we want a list in this form: [a,b,X|Y]. The following definiton tries to follow this aim.
list_heads(T) --> term(T).
list_heads(Heads) --> term(T), comma, list_heads(Ts), {Heads =..[',',T,Ts]}.


left_bracket --> one_time(eat("["),'[').
right_bracket --> one_time(eat("]"),']').

% an atom is a lowercase letter and the next letter is not a letter of type alnum (lower, uppercase or digit)
atom(Name) --> lowercase(Code), next_letter(L), {not(code_type(L,alnum)),atom_codes(Name,[Code])}.

% an atom is a string beginning with a lowercase letter.
atom(Name) --> lowercase(Code), nameChars(Codes),{atom_codes(Name,[Code|Codes])}.



constant(Number) --> integer(Digits),{number_codes(Number,Digits)}.
constant(Name) --> atom(Name).

integer([Digit]) --> digit(Digit), next_letter(L), {not(code_type(L,digit))}.
integer([Digit|Digits]) --> digit(Digit), integer(Digits).

% Variable is a letter (upper-  or  lowercase) and next letter is not of code_type csym.
% code_type(L,csym) : Char  is  a  letter   (upper-  or  lowercase),  digit   or  the
%         underscore  (_).
variable(X) --> uppercase(X1), next_letter(L), {not(code_type(L,csym)),atom_codes(X,[X1])}.

% Variable is a string beginning with an uppercase letter.
variable(X) --> uppercase(X1), nameChars(X2), {atom_codes(X,[X1|X2])}.

uppercase(X) --> eat([X]),{code_type(X,upper), X=<127}.
lowercase(X) --> eat([X]),{code_type(X,lower), X=< 127}.
digit(X) --> eat([X]),{code_type(X,digit)}.


nameChar(X)-->eat([X]),{code_type(X,csym)}.
nameChars([X]) --> nameChar(X), next_letter(L), {not(code_type(L,csym))}.
nameChars([X|Xs]) --> nameChar(X), nameChars(Xs).

% ----auxiliary predicates-------%
% removes the codes in String from the remaining codes
eat(String)-->modify(ids_state^state2-parser_state^string,Before,After),
	{append(String,After,Before)}.

% looks at the next letter, but does not consume it
next_letter(L)-->getf(ids_state^state2-parser_state^string,String),
	{String=[L|_]}.
next_letter(0)-->getf(ids_state^state2-parser_state^string,[]).


%= end (*)
%% generates an error if we have a part of the string unparsed.
end-->modify(ids_state^state2-parser_state^string,[],[]).
end-->modify(ids_state^state2-parser_state^string,Rest,[]),
	{length(Rest,N)},
	error(remaining(Rest),15+N).

one_time(X,_)-->call(X).
one_time(_,What)-->error(missing(What),5).
one_time(X,What)-->error(superfluous(What),5),call(X),one_time(X,What).

% elements of a list have the form [Term, Error]
% will be changed to 2 lists. One is a list of term
% the other one is a list of errors

seperateTerm_Errorlist([[Term,N-X]|Rest],[Term|TermRest],[N-X-Term|ErrorlistRest]):-
	X\=[],
	seperateTerm_Errorlist(Rest,TermRest,ErrorlistRest).
seperateTerm_Errorlist([[Term,_-X]|Rest],[Term|TermRest],ErrorlistRest):-
	X=[],
	seperateTerm_Errorlist(Rest,TermRest,ErrorlistRest).

seperateTerm_Errorlist([],[],[]).

% remove the spaces tha end a string 
remove_last_space_in_list(List,ResultList):-
	is_last_element_space_or_cr(List),
	remove_last_element(List,NewList),
	remove_last_space_in_list(NewList,ResultList).
remove_last_space_in_list(List,List):-not(is_last_element_space_or_cr(List)).

% remove the spaces that begin the string
remove_first_space_in_list([X|Rest],ResultList):-
	X=32,
	remove_first_element([X|Rest],NewList),
	remove_first_space_in_list(NewList,ResultList).
remove_first_space_in_list([X|Rest],[X|Rest]):-not(X=32).

% test if the last element of a string is a space
is_last_element_space_or_cr(List):-
	last(List,Element), 
	(   Element=32;Element=13)
	.





