%% Testing of recursive predicates
%% Author: Yannick Versley
%% last changed: $Id: test_recursive.pl,v 1.5 2004/03/29 07:44:57 nthinh Exp $

/**
test_with_goal testet zwei Prdikate mit einer Anfrage, indem es jeweils
dprove mit der Anfrage und dem Prdikat aufruft.

dprove simuliert die Ausfhrung einer Anfrage beim Testen eines Prdikats.
Mit
?- dprove(peano(X),X,[peano(0)-[],peano(s(X1))-[peano(X1)]],2,Result).
sollte dann Result der Reihe nach mit
success(0), success(s(0)), success(s(0)), depth_reached(peano(s(_G316)))
unifiziert werden.

Danach wird mit find_mismatch geschaut, ob es
Ergebnisse gibt, die in das eine Prdikat findet oder das andere nicht
(dann wird das Ergebnis als false_positive(Result) ausgegeben), oder
umgekehrt (dann wird das Ergebnis als false_negative(Result) ausgegeben.

Wenn das eine Prdikat immer terminiert hat, aber das andere nicht, dann
wird ein termination_problem(Goal) ausgegeben.

Die Alternative hierzu wre etwa, die Prdikatsdefinitionen in eine Datei
zu schreiben, die dann einzulesen und mit call(Anfrage) oder
call_with_depth_limit(Anfrage,Limit) zu testen und uninstantiierte Aufrufe
(X is Y+Z, Y uninstantiiert) und hnliche Probleme mit catch abzufangen.

*/

:- module(test_recursive,[dprove/5,
			 test_with_goal/4]).

:- use_module(ids).
:- use_module(logging).

builtin(append(_,_,_)).
builtin(_ \= _).
builtin(_=_).

% dprove(Goal,Vars
% a depth-limited prover
% Goal:    the goal we want to test.
%	   this can be a subgoal
% Vars:	   what we want to have in the goal
% Clauses: the clauses of the predicate in Head-[Bodylist] form
%          the variables have to be left as variables.
% MaxDepth:the depth that is left
dprove(Goal,Vars,Clauses,MaxDepth,Result):-
	logging(test_recursive,'<dproveA>Input: Goal=~w, Vars=~w, Clauses=~w, MaxDepth=~w',[Goal,Vars,Clauses,MaxDepth]),
	not(builtin(Goal)),!,
	logging(test_recursive,'<dproveC>Goal is not a builtin predicate.',[]),
	(   succ(MD1,MaxDepth)->
	% this is inefficient, but we don't want to modify the clauses
	% of the program by unifying one of them with the goal.
	(copy_term(Clauses,Clauses1),
	 % to prove a goal, choose a clause from Clauses whose head unifies with the goal.
	    (member(Goal-Body,Clauses1)*->
	       (   logging(test_recursive,'<dproveA>Goal-Body from Clauses1=~w-~w',[Goal,Body]),
		   dprove1(Body,Vars,Clauses,MD1,Result)
		   %logging(test_recursive,'<dprove>proved body=~w, Result=~w',[Body,Result])
	       );
	       Result=failed_because(Vars,Goal)
	      % logging(test_recursive,'<dprove>Result=~w',[Result])
	    )
        );
        Result=depth_reached(Goal)
	),
        logging(test_recursive,'<dproveA>Result=~w',[Result]).
dprove(Var is Expr,Vars,_Clauses,_MD1,Result):-	
	logging(test_recursive,'<dproveB>Input: Var is Expr=~w is ~w, Vars=~w',[Var,Expr,Vars]),
	ground(Expr)->
	   (   Var is Expr,
	       Result=success(Vars)
	   );
	Result=not_instantiated(Var is Expr),
	logging(test_recursive,'<dproveB>Output: Result=~w',[Result]).
dprove(Goal,Vars,_Clauses,_MD1,Result):-
	logging(test_recursive,'<dproveC>Input: Goal=~w, Vars=~w',[Goal,Vars]),
	builtin(Goal),!,
	logging(test_recursive,'<dproveC>Goal is a builtin predicate. call(~w) will be executed',[Goal]),
	call(Goal), 
	Result=success(Vars),
	%            Result=failed_because(Goal),
	logging(test_recursive,'<dproveC>Output: Result=~w',[Result]).

% go through the subgoals we have left.
dprove1([],Vars,_,_,success(Vars)):-
	logging(test_recursive,'<dprove1a>Subgoal list is empty. Output: RESULT=success(~w)',[Vars]).
dprove1([Goal|Rest],Vars,Clauses,MaxDepth,Result):-
	logging(test_recursive,'<dprove1b>Input: subgoals [Goal|Rest]=[~w|~w], Vars=~w',[Goal,Rest,Vars]),
	logging(test_recursive,'<dprove1b>Input: Clauses=~w, MaxDepth=~w',[Clauses,MaxDepth]),
	assume(not(var(Rest))),
	% logging(test_recursive,'<dprove1b>Prove subgoal=~w',[Goal]),
	copy_term(Clauses,Clauses1),
	dprove(Goal,Vars,Clauses1,MaxDepth,Result1),
	logging(test_recursive,'<dprove1b>Result1=~w',[Result1]),
	(Result1=success(_)->
	dprove1(Rest,Vars,Clauses,MaxDepth,Result);
	Result=Result1),
	logging(test_recursive,'<dprove1b>Output: Result=~w',[Result]).

% test_with_goal(Clauses1, Clauses2, Goal, Errors)
% compares the solution sets of two predicates
% Clauses1: the clauses of the sample predicates
% Clauses2: the clauses of the user predicates
%	    these two have to be lists of clauses in
%	    the form Head-[BodyGoals]
test_with_goal(Clauses1, Clauses2, Goal, Errors):-
	logging(test_recursive,'<test_with_goal>Input: Clauses1=~w, Clauses2=~w, Goal=~w',[Clauses1,Clauses2,Goal]),
	findall(Res,(logging(test_recursive,'<test_with_goal>Prove the SAMPLE predicate: Goal=~w, Clauses1:~w',[Goal,Clauses1]),
		     dprove(Goal,Goal,Clauses1,10,Res)),
		SolutionSet1),
	findall(Res,(logging(test_recursive,'<test_with_goal>Prove the USER predicate: Goal=~w, Clauses2:~w',[Goal,Clauses2]),
		     dprove(Goal,Goal,Clauses2,10,Res)),
		SolutionSet2),
	logging(test_recursive,'<test_with_goal>SolutionSet1:~w',[SolutionSet1]),
	logging(test_recursive,'<test_with_goal>SolutionSet2:~w',[SolutionSet2]),
	findall(Wrong,find_mismatch(Wrong,SolutionSet1,SolutionSet2),
		Errors),
	logging(test_recursive,'<test_with_goal>Errors after test: ~w',[Errors]).

subsumes(X,Y):-
	copy_term(X,X1),
	X1=Y,
	X1==X.

find_mismatch(termination_problem(Goal),Set1,Set2):-
	member(depth_reached(Goal),Set2),
	not(member(depth_reached(_),Set1)).

% USER solution has more successful uoutputs than SAMPLE solution.
find_mismatch(false_positive(X),Set1,Set2):-
	member(success(X),Set2),
	not((member(success(X1),Set1),X=@=X1)),
	not(member(depth_reached(_),Set1)).
% SAMPLE solution has more successful outputs than USER solution.
find_mismatch(missing_solution(X),Set1,Set2):-
	member(success(X),Set1),
	not((member(success(X1),Set2),X=@=X1)),
	not(member(depth_reached(_),Set2)).
find_mismatch(aplly_wrong_builtin_predicate(Wrong,Right),Set1,Set2):-
	member(failed_because(X),Set1),
	member(failed_because(Y),Set2),
%	X=@=Y,
	logging(test_recursive,'<find_mismatch>X=~w, Y=~w',[X,Y]),
	not(member(depth_reached(_),Set1)),
	not(member(depth_reached(_),Set2)),
	X=..[Right|_],
	Y=..[Wrong|_],
	logging(test_recursive,'<find_mismatch>Right=~w, Wrong=~w',[Right,Wrong]),
%	wildcard_match(Right,Wrong),
	logging(test_recursive,'<find_mismatch>end.',[]).


test(1,Errors):-
	test_with_goal([rev([],X1,X1)-[],
			rev([A2|Rest2],Accu2,Result2)-
			[rev(Rest2,[A2|Accu2],Result2)]],
		       [rev(_,X1,X1)-[],
			rev([A2|Rest2],Accu2,[A2|Result2])-
			[rev([A2|Rest2],[A2|Accu2],Result2)]],
		       rev([1,2,3],[],_X),
		       Errors).
test(2,Errors):-
	test_with_goal([rev([],X1,X1)-[],
			rev([A2|Rest2],Accu2,Result2)-
			[rev(Rest2,[A2|Accu2],Result2)]],
		       [rev([],[],X1)-[],
			rev([A2|Rest2],Accu2,Result2)-
			[rev(Rest2,[A2|Accu2],Result2)]],
		       rev([1,2,3],[],_X),
		       Errors).
			

